#
# Autogenerated by Thrift
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#

from thrift.Thrift import *

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class Status(object):
  DEAD = 0
  STARTING = 1
  ALIVE = 2
  STOPPING = 3
  STOPPED = 4
  WARNING = 5

  _VALUES_TO_NAMES = {
    0: "DEAD",
    1: "STARTING",
    2: "ALIVE",
    3: "STOPPING",
    4: "STOPPED",
    5: "WARNING",
  }

  _NAMES_TO_VALUES = {
    "DEAD": 0,
    "STARTING": 1,
    "ALIVE": 2,
    "STOPPING": 3,
    "STOPPED": 4,
    "WARNING": 5,
  }


class Endpoint(object):
  """
  Attributes:
   - host
   - port
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'host', None, None, ), # 1
    (2, TType.I32, 'port', None, None, ), # 2
  )

  def __init__(self, host=None, port=None,):
    self.host = host
    self.port = port

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.host = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.port = iprot.readI32();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Endpoint')
    if self.host != None:
      oprot.writeFieldBegin('host', TType.STRING, 1)
      oprot.writeString(self.host)
      oprot.writeFieldEnd()
    if self.port != None:
      oprot.writeFieldBegin('port', TType.I32, 2)
      oprot.writeI32(self.port)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()
    def validate(self):
      return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ServiceInstance(object):
  """
  Attributes:
   - serviceEndpoint
   - additionalEndpoints
   - status
   - shard
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'serviceEndpoint', (Endpoint, Endpoint.thrift_spec), None, ), # 1
    (2, TType.MAP, 'additionalEndpoints', (TType.STRING,None,TType.STRUCT,(Endpoint, Endpoint.thrift_spec)), None, ), # 2
    (3, TType.I32, 'status', None, None, ), # 3
    (4, TType.I32, 'shard', None, None, ), # 4
  )

  def __init__(self, serviceEndpoint=None, additionalEndpoints=None, status=None, shard=None,):
    self.serviceEndpoint = serviceEndpoint
    self.additionalEndpoints = additionalEndpoints
    self.status = status
    self.shard = shard

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.serviceEndpoint = Endpoint()
          self.serviceEndpoint.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.MAP:
          self.additionalEndpoints = {}
          (_ktype1, _vtype2, _size0 ) = iprot.readMapBegin() 
          for _i4 in xrange(_size0):
            _key5 = iprot.readString();
            _val6 = Endpoint()
            _val6.read(iprot)
            self.additionalEndpoints[_key5] = _val6
          iprot.readMapEnd()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.status = iprot.readI32();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I32:
          self.shard = iprot.readI32();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ServiceInstance')
    if self.serviceEndpoint != None:
      oprot.writeFieldBegin('serviceEndpoint', TType.STRUCT, 1)
      self.serviceEndpoint.write(oprot)
      oprot.writeFieldEnd()
    if self.additionalEndpoints != None:
      oprot.writeFieldBegin('additionalEndpoints', TType.MAP, 2)
      oprot.writeMapBegin(TType.STRING, TType.STRUCT, len(self.additionalEndpoints))
      for kiter7,viter8 in self.additionalEndpoints.items():
        oprot.writeString(kiter7)
        viter8.write(oprot)
      oprot.writeMapEnd()
      oprot.writeFieldEnd()
    if self.status != None:
      oprot.writeFieldBegin('status', TType.I32, 3)
      oprot.writeI32(self.status)
      oprot.writeFieldEnd()
    if self.shard != None:
      oprot.writeFieldBegin('shard', TType.I32, 4)
      oprot.writeI32(self.shard)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()
    def validate(self):
      return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



from functools import partial
import itertools
import posixpath
import threading

try:
  from twitter.common import log
except ImportError:
  import logging as log

from twitter.common.concurrent import Future

from .group_base import (
    Capture,
    GroupBase,
    GroupInterface,
    Membership,
    set_different)

from kazoo.client import KazooClient
from kazoo.protocol.states import (
    EventType,
    KazooState,
    KeeperState)

import kazoo.security as ksec
import kazoo.exceptions as ke


# TODO(wickman) Put this in twitter.common somewhere?
def partition(items, predicate=bool):
  a, b = itertools.tee((predicate(item), item) for item in items)
  return ([item for pred, item in a if not pred], [item for pred, item in b if pred])


class KazooGroup(GroupBase, GroupInterface):
  """
    An implementation of GroupInterface against Kazoo.
  """
  DISCONNECT_EXCEPTIONS = (ke.ConnectionLoss, ke.OperationTimeoutError, ke.SessionExpiredError)

  @classmethod
  def translate_acl(cls, acl):
    if not isinstance(acl, dict) or any(key not in acl for key in ('perms', 'scheme', 'id')):
      raise TypeError('Expected acl to be Acl-like, got %s' % type(acl))
    return ksec.ACL(acl['perms'], ksec.Id(acl['scheme'], acl['id']))

  @classmethod
  def translate_acl_list(cls, acls):
    if acls is None:
      return acls
    try:
      acls = list(acls)
    except (ValueError, TypeError):
      raise TypeError('ACLs should be a list, got %s' % type(acls))
    if all(isinstance(acl, ksec.ACL) for acl in acls):
      return acls
    else:
      return [cls.translate_acl(acl) for acl in acls]

  def __init__(self, zk, path, acl=None):
    if not isinstance(zk, KazooClient):
      raise TypeError('KazooGroup must be initialized with a KazooClient')
    self._zk = zk
    self.__state = zk.state
    self.__listener_queue = []
    self.__queue_lock = threading.Lock()
    self._zk.add_listener(self.__state_listener)
    self._path = '/' + '/'.join(filter(None, path.split('/')))  # normalize path
    self._members = {}
    self._member_lock = threading.Lock()
    self._acl = self.translate_acl_list(acl)

  def __state_listener(self, state):
    """Process appropriate callbacks on any kazoo state transition."""
    with self.__queue_lock:
      self.__state = state
      self.__listener_queue, triggered = partition(self.__listener_queue,
          lambda element: element[0] == state)
    for _, callback in triggered:
      callback()

  def _once(self, keeper_state, callback):
    """Ensure a callback is called once we reach the given state: either
       immediately, if currently in that state, or on the next transition to
       that state."""
    invoke = False
    with self.__queue_lock:
      if self.__state != keeper_state:
        self.__listener_queue.append((keeper_state, callback))
      else:
        invoke = True
    if invoke:
      callback()

  def __on_connected(self, callback):
    return self.__on_state(callback, KazooState.CONNECTED)

  def __on_expired(self, callback):
    return self.__on_state(callback, KazooState.LOST)

  def info(self, member, callback=None):
    if member == Membership.error():
      raise self.InvalidMemberError('Cannot get info on error member!')

    capture = Capture(callback)

    def do_info():
      self._zk.get_async(path).rawlink(info_completion)

    with self._member_lock:
      member_future = self._members.setdefault(member, Future())

    member_future.add_done_callback(lambda future: capture.set(future.result()))

    dispatch = False
    with self._member_lock:
      if not member_future.done() and not member_future.running():
        try:
          dispatch = member_future.set_running_or_notify_cancel()
        except:
          pass

    def info_completion(result):
      try:
        content, stat = result.get()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, do_info)
        return
      except ke.NoNodeException:
        future = self._members.pop(member, Future())
        future.set_result(Membership.error())
        return
      except ke.KazooException as e:
        log.warning('Unexpected Kazoo result in info: (%s)%s' % (type(e), e))
        future = self._members.pop(member, Future())
        future.set_result(Membership.error())
        return
      self._members[member].set_result(content)

    if dispatch:
      path = posixpath.join(self._path, self.id_to_znode(member.id))
      do_info()

    return capture()

  def join(self, blob, callback=None, expire_callback=None):
    membership_capture = Capture(callback)
    expiry_capture = Capture(expire_callback)

    def do_join():
      self._zk.create_async(
          path=posixpath.join(self._path, self.MEMBER_PREFIX),
          value=blob,
          acl=self._acl,
          sequence=True,
          ephemeral=True,
          makepath=True
      ).rawlink(acreate_completion)

    def do_exists(path):
      self._zk.exists_async(path, watch=exists_watch).rawlink(partial(exists_completion, path))

    def exists_watch(event):
      if event.type == EventType.DELETED:
        expiry_capture.set()

    def expire_notifier():
      self._once(KazooState.LOST, expiry_capture.set)

    def exists_completion(path, result):
      try:
        if result.get() is None:
          expiry_capture.set()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, partial(do_exists, path))

    def acreate_completion(result):
      try:
        # TODO(wickman) Kazoo has a bug:
        #    https://github.com/python-zk/kazoo/issues/106
        #    https://github.com/python-zk/kazoo/pull/107
        # Remove this one 1.3 is cut.
        path = self._zk.unchroot(result.get())
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, do_join)
        return
      except ke.KazooException as e:
        log.warning('Unexpected Kazoo result in join: (%s)%s' % (type(e), e))
        membership = Membership.error()
      else:
        created_id = self.znode_to_id(path)
        membership = Membership(created_id)
        with self._member_lock:
          result_future = self._members.get(membership, Future())
          result_future.set_result(blob)
          self._members[membership] = result_future
        if expire_callback:
          self._once(KazooState.CONNECTED, expire_notifier)
          do_exists(path)

      membership_capture.set(membership)

    do_join()
    return membership_capture()

  def cancel(self, member, callback=None):
    capture = Capture(callback)

    def do_cancel():
      self._zk.delete_async(posixpath.join(self._path, self.id_to_znode(member.id))).rawlink(
          adelete_completion)

    def adelete_completion(result):
      try:
        success = result.get()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, do_cancel)
        return
      except ke.NoNodeError:
        success = True
      except ke.KazooException as e:
        log.warning('Unexpected Kazoo result in cancel: (%s)%s' % (type(e), e))
        success = False

      future = self._members.pop(member.id, Future())
      future.set_result(Membership.error())
      capture.set(success)

    do_cancel()
    return capture()

  def monitor(self, membership=frozenset(), callback=None):
    capture = Capture(callback)

    def wait_exists():
      self._zk.exists_async(self._path, exists_watch).rawlink(exists_completion)

    def exists_watch(event):
      if event.state == KeeperState.EXPIRED_SESSION:
        wait_exists()
        return
      if event.type == EventType.CREATED:
        do_monitor()
      elif event.type == EventType.DELETED:
        wait_exists()

    def exists_completion(result):
      try:
        stat = result.get()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, wait_exists)
        return
      except ke.NoNodeError:
        wait_exists()
        return
      except ke.KazooException as e:
        log.warning('Unexpected exists_completion result: (%s)%s' % (type(e), e))
        return

      if stat:
        do_monitor()

    def do_monitor():
      self._zk.get_children_async(self._path, get_watch).rawlink(get_completion)

    def get_watch(event):
      if event.state == KeeperState.EXPIRED_SESSION:
        wait_exists()
        return
      if event.state != KeeperState.CONNECTED:
        return
      if event.type == EventType.DELETED:
        wait_exists()
        return
      if event.type != EventType.CHILD:
        return
      if set_different(capture, membership, self._members):
        return
      do_monitor()

    def get_completion(result):
      try:
        children = result.get()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, do_monitor)
        return
      except ke.NoNodeError:
        wait_exists()
        return
      except ke.KazooException as e:
        log.warning('Unexpected get_completion result: (%s)%s' % (type(e), e))
        capture.set(set([Membership.error()]))
        return
      self._update_children(children)
      set_different(capture, membership, self._members)

    do_monitor()
    return capture()

  def list(self):
    wait_event = threading.Event()
    while True:
      wait_event.clear()
      try:
        try:
          return sorted(Membership(self.znode_to_id(znode))
                        for znode in self._zk.get_children(self._path)
                        if self.znode_owned(znode))
        except ke.NoNodeException:
          return []
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, wait_event.set)
        wait_event.wait()


class ActiveKazooGroup(KazooGroup):
  def __init__(self, *args, **kwargs):
    super(ActiveKazooGroup, self).__init__(*args, **kwargs)
    self._monitor_queue = []
    self._monitor_members()

  def monitor(self, membership=frozenset(), callback=None):
    capture = Capture(callback)
    if not set_different(capture, membership, self._members):
      self._monitor_queue.append((membership, capture))

    return capture()

  def _monitor_members(self):
    def wait_exists():
      self._zk.exists_async(self._path, exists_watch).rawlink(exists_completion)

    def exists_watch(event):
      if event.state == KeeperState.EXPIRED_SESSION:
        wait_exists()
        return
      if event.type == EventType.CREATED:
        do_monitor()
      elif event.type == EventType.DELETED:
        wait_exists()

    def exists_completion(result):
      try:
        stat = result.get()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, wait_exists)
        return
      except ke.NoNodeError:
        wait_exists()
        return
      except ke.KazooException as e:
        log.warning('Unexpected exists_completion result: (%s)%s' % (type(e), e))
        return

      if stat:
        do_monitor()

    def do_monitor():
      self._zk.get_children_async(self._path, get_watch).rawlink(get_completion)

    def get_watch(event):
      if event.state == KeeperState.EXPIRED_SESSION:
        wait_exists()
        return
      if event.state != KeeperState.CONNECTED:
        return
      if event.type == EventType.DELETED:
        wait_exists()
        return

      do_monitor()

    def get_completion(result):
      try:
        children = result.get()
      except self.DISCONNECT_EXCEPTIONS:
        self._once(KazooState.CONNECTED, do_monitor)
        return
      except ke.NoNodeError:
        wait_exists()
        return
      except ke.KazooException as e:
        log.warning('Unexpected get_completion result: (%s)%s' % (type(e), e))
        return

      children = [child for child in children if self.znode_owned(child)]
      _, new = self._update_children(children)
      for child in new:
        def devnull(*args, **kw): pass
        self.info(child, callback=devnull)

      monitor_queue = self._monitor_queue[:]
      self._monitor_queue = []
      members = set(Membership(self.znode_to_id(child)) for child in children)
      for membership, capture in monitor_queue:
        if set(membership) != members:
          capture.set(members)
        else:
          self._monitor_queue.append((membership, capture))

    do_monitor()



try:
  from twitter.common import log
except ImportError:
  import logging as log

from twitter.common.zookeeper.group.group_base import GroupInterface

try:
  from twitter.common.zookeeper.client import ZooKeeper
  from twitter.common.zookeeper.group.group import (
      ActiveGroup,
      Group)
  def pick_zkpython_group(zk, on_join, on_leave):
    # The default underlying implementation is Group if no active monitoring
    # is requested of the ServerSet.  If active monitoring is requested by
    # on_join or on_leave, then use ActiveGroup by default, which has better
    # performance on monitor/iter calls.
    if isinstance(zk, ZooKeeper):
      return Group if (on_join is None and on_leave is None) else ActiveGroup
except ImportError as e:
  def pick_zkpython_group(zk, on_join, on_leave):
    return None

try:
  from kazoo.client import KazooClient
  from twitter.common.zookeeper.group.kazoo_group import ActiveKazooGroup, KazooGroup
  def pick_kazoo_group(zk, on_join, on_leave):
    if isinstance(zk, KazooClient):
      return KazooGroup if (on_join is None and on_leave is None) else ActiveKazooGroup
except ImportError as e:
  def pick_kazoo_group(zk, on_join, on_leave):
    return None

GROUP_SELECTORS = [pick_zkpython_group, pick_kazoo_group]

from .endpoint import ServiceInstance


def first(iterable):
  for element in iterable:
    if element:
      return element


def validate_group_implementation(underlying):
  assert issubclass(underlying, GroupInterface), (
    'Underlying group implementation must be a subclass of GroupInterface, got %s'
    % type(underlying))


class ServerSet(object):
  """
    A dynamic set of service endpoints tracked by Zookeeper.
  """

  def __init__(self, zk, path, underlying=None, on_join=None, on_leave=None, **kwargs):
    """
      Construct a ServerSet at :path given zookeeper handle :zh.

      If :underlying is specified, use that as the underlying Group implementation.  Must be a
      subclass of twitter.common.zookeeper.group.GroupInterface.

      If :on_join is specified, it will be called with a ServiceInstance object every time a
      new service joins the ServerSet.  If :on_leave is specified, it will be called with
      a ServiceInstance object every time a server leaves the ServerSet.

      All remaining arguments are passed to the underlying Group implementation.
    """
    underlying = underlying or first(
        pick_group(zk, on_join, on_leave) for pick_group in GROUP_SELECTORS)
    if underlying is None:
      raise ValueError("Couldn't find a suitable group implementation!")

    validate_group_implementation(underlying)

    self._path = path
    self._group = underlying(zk, path, **kwargs)
    def devnull(*args, **kw): pass
    self._on_join = on_join or devnull
    self._on_leave = on_leave or devnull
    self._members = {}
    if on_join or on_leave:
      self._internal_monitor(set(self._members))

  def join(self, endpoint, additional=None, shard=None, callback=None, expire_callback=None):
    """
      Given 'endpoint' (twitter.common.zookeeper.serverset.Endpoint) and an
      optional map 'additional' of string => endpoint (also Endpoint), and an
      optional shard id, create a ServiceInstance and join it into this ServerSet.

      If 'callback' is provided, the join will be done asynchronously and
      'callback' will be called with the Membership object associated with
      the ServiceInstance.  If joining fails, Membership.error() will be
      returned.  If 'callback' is not provided, join will return with this
      information synchronously.

      If 'expire_callback' is provided, it will be called if the membership
      is severed for any reason such as session expiration or malice.
    """
    service_instance = ServiceInstance.pack(ServiceInstance(endpoint, additional, shard=shard))
    return self._group.join(service_instance, callback=callback, expire_callback=expire_callback)

  def cancel(self, membership, callback=None):
    """Cancel membership in the ServerSet."""
    return self._group.cancel(membership, callback=callback)

  def __iter__(self):
    """Iterate over the services (ServiceInstance objects) in this ServerSet."""
    for member in self._group.list():
      try:
        yield ServiceInstance.unpack(self._group.info(member))
      except Exception as e:
        log.warning('Failed to deserialize endpoint: %s' % e)
        continue

  def _internal_monitor(self, members):
    cached = set(self._members)
    new_members = members - cached
    old_members = cached - members

    for service_instance in map(self._members.pop, old_members):
      self._on_leave(service_instance)

    def make_callback(member_id):
      def callback(service_instance):
        try:
          self._members[member_id] = ServiceInstance.unpack(service_instance)
        except Exception as e:
          log.warning('Failed to deserialize endpoint: %s' % e)
          return
        self._on_join(self._members[member_id])

      return callback

    for member_id in new_members:
      self._group.info(member_id, make_callback(member_id))

    self._group.monitor(members, self._internal_monitor)



from abc import abstractmethod
import posixpath
import threading

from twitter.common.concurrent import Future
from twitter.common.lang import Interface


class Membership(object):
  ERROR_ID = -1

  @staticmethod
  def error():
    return Membership(_id=Membership.ERROR_ID)

  def __init__(self, _id):
    self._id = _id

  @property
  def id(self):
    return self._id

  def __lt__(self, other):
    return self._id < other._id

  def __eq__(self, other):
    if not isinstance(other, Membership):
      return False
    return self._id == other._id

  def __ne__(self, other):
    return not self == other

  def __hash__(self):
    return hash(self._id)

  def __repr__(self):
    if self._id == Membership.ERROR_ID:
      return 'Membership.error()'
    else:
      return 'Membership(%r)' % self._id


class GroupInterface(Interface):
  """
    A group of members backed by immutable blob data.
  """

  @abstractmethod
  def join(self, blob, callback=None, expire_callback=None):
    """
      Joins the Group using the blob.  Returns Membership synchronously if
      callback is None, otherwise returned to callback.  Returns
      Membership.error() on failure to join.

      If expire_callback is provided, it is called with no arguments when
      the membership is terminated for any reason.
    """

  @abstractmethod
  def info(self, membership, callback=None):
    """
      Given a membership, return the blob associated with that member or
      Membership.error() if no membership exists.  If callback is provided,
      this operation is done asynchronously.
    """

  @abstractmethod
  def cancel(self, membership, callback=None):
    """
      Cancel a given membership.  Returns true if/when the membership does not
      exist.  Returns false if the membership exists and we failed to cancel
      it.  If callback is provided, this operation is done asynchronously.
    """

  @abstractmethod
  def monitor(self, membership_set=frozenset(), callback=None):
    """
      Given a membership set, return once the underlying group membership is
      different.  If callback is provided, this operation is done
      asynchronously.
    """

  @abstractmethod
  def list(self):
    """
      Synchronously return the list of underlying members.  Should only be
      used in place of monitor if you cannot afford to block indefinitely.
    """


# TODO(wickman) The right abstraction here is probably IAsyncResult from Kazoo.
# Kill this in favor of the better abstraction.
class Capture(object):
  """
    A Capture is a mechanism to capture a value to be dispatched via a
    callback or blocked upon.  If Capture is supplied with a callback, the
    callback is called once the value is available, in which case
    Capture.__call__() will return immediately.  If no callback has been
    supplied, Capture.__call__() blocks until a value is available.
  """
  def __init__(self, callback=None):
    self._value = None
    self._event = threading.Event()
    self._callback = callback

  def set(self, value=None):
    self._value = value
    self._event.set()
    if self._callback:
      if value is not None:
        self._callback(value)
      else:
        self._callback()
      self._callback = None

  def get(self):
    self._event.wait()
    return self._value

  def __call__(self):
    if self._callback:
      return None
    return self.get()


def set_different(capture, current_members, actual_members):
  current_members = set(current_members)
  actual_members = set(actual_members)
  if current_members != actual_members:
    capture.set(actual_members)
    return True


class GroupBase(object):
  class GroupError(Exception): pass
  class InvalidMemberError(GroupError): pass

  MEMBER_PREFIX = 'member_'

  @classmethod
  def znode_owned(cls, znode):
    return posixpath.basename(znode).startswith(cls.MEMBER_PREFIX)

  @classmethod
  def znode_to_id(cls, znode):
    znode_name = posixpath.basename(znode)
    assert znode_name.startswith(cls.MEMBER_PREFIX)
    return int(znode_name[len(cls.MEMBER_PREFIX):])

  @classmethod
  def id_to_znode(cls, _id):
    return '%s%010d' % (cls.MEMBER_PREFIX, _id)

  def __iter__(self):
    return iter(self._members)

  def __getitem__(self, member):
    return self.info(member)

  def _update_children(self, children):
    """
      Given a new child list [znode strings], return a tuple of sets of Memberships:
        left: the children that left the set
        new: the children that joined the set
    """
    cached_children = set(self._members)
    current_children = set(Membership(self.znode_to_id(znode))
        for znode in filter(self.znode_owned, children))
    new = current_children - cached_children
    left = cached_children - current_children
    for child in left:
      future = self._members.pop(child, Future())
      future.set_result(Membership.error())
    for child in new:
      self._members[child] = Future()
    return left, new




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(src/python/twitter/common/zookeeper/BUILD:zookeeper-packaged)

from setuptools import setup

setup(**
{   'description': "Implementations of Twitter's service discovery libraries on top of Kazoo.",
    'install_requires': [   'kazoo==1.3.1',
                            'twitter.common.concurrent==0.1.0',
                            'twitter.common.metrics==0.1.0'],
    'name': 'twitter.common.zookeeper',
    'namespace_packages': [   'gen',
                              'gen.twitter',
                              'gen.twitter.thrift',
                              'twitter',
                              'twitter.common'],
    'package_data': {   },
    'package_dir': {   '': 'src'},
    'packages': [   '.',
                    'gen',
                    'gen.twitter',
                    'gen.twitter.thrift',
                    'gen.twitter.thrift.endpoint',
                    'twitter',
                    'twitter.common',
                    'twitter.common.zookeeper',
                    'twitter.common.zookeeper.group',
                    'twitter.common.zookeeper.serverset'],
    'version': '0.1.0'}
)
